;tsp-mit-BS.scm
(load "../../breakpoint.scm")

; verwendet werden:
(load "gallenbacher-graph-kanten.scm")
; Verwendet wird davon die Kantenliste "kanten" ((A B 69) ...)
; und die Zugriffsfunktion (nachfolge-kanten <aktuelle-stadt> <kanten>)
(load "wege.scm")
; mit den Funktionen (neuer-weg <weg> <kante>)
; und (alle-nachfolge-wege <weg> <kanten>) 
; dabei muss hier beachtet werden, dass davon kein Zyklus gebildet wird, der
; Rueckweg zum Start also gesondert eingebunden werden muss
; Wege sind Listen aus der Laenge und den besuchten Knoten (rueckwaerts)
; Muster: (60 X B I)


;;; ===== Anforderung TSP =====
; Jeder Ort wird bei einer Rundreise durch alle Orte genau einmal angefahren.
; Ziel ist, die dabei gefahrene Strecke zu optimieren.
; Wegen der Optimierung (also kein Satisfizierungsproblem im eigentlichen Sinn)
; muss eine Liste aller gefundenen Loesungen oder die bis dahin optimale geführt werden.
; Das legt ein endrekursives Vorgehen nahe.
; Da alle Loesungen gleich viele Kanten enthalten muessen, bietet sich die 
; Breitensuche als Verfahren an.

;;; ===== Hilfsfunktion anzahl-knoten =====
; Voraussetzung: Die Liste kanten enthaelt jeweils
; auch die Rueckkante
(define
  (anzahl-knoten kanten)
  (define
    (anzahl-intern kanten gefunden)
    ; arbeitet endrekursiv
    (cond
      ((null? kanten) (length gefunden))
      ((member (first (first kanten)) gefunden)
       (anzahl-intern (rest kanten) gefunden))
      (else
       (anzahl-intern
        (rest kanten)
        (cons (first (first kanten)) gefunden)))))
  (anzahl-intern kanten '()))
; Test:
;(anzahl-knoten kanten)

;;; ===== Hilfsfunktion knoten-im-weg =====
; bestimmt die Anzahl der Knoten in weg
(define
  (knoten-im-weg weg)
  (length (rest weg)))
; Test:
;(knoten-im-weg '(64 H B I))

;;; ===== Hilfsfunktion nachbar-staedte =====
; bestimmt alle Nachbarstaedte der aktuellen Stadt
(define
  (nachbar-staedte aktuelle-stadt kanten)
  (map second (nachfolge-kanten aktuelle-stadt kanten)))
;Test:
;(nachbar-staedte 'I kanten)
;;; ===== Hilfsfunktion besserer-weg =====
; bestimmt den kuerzeren zweier Wege
(define
  (besserer-weg weg-1 weg-2)
  (if (< (first weg-1) (first weg-2)) weg-1 weg-2))
;Test:
;(besserer-weg '(64 H B I) '(+inf.0 I))

;;; ===== Hilfsfunktion zum-start =====
; gibt zu einem Weg, von dem es einen Rueckweg zum Start gibt,
; diesen zurueck
(define
  (zum-start weg)
  (define
    (intern die-kanten start)
    (if
     (equal? start (second (first die-kanten)))
     (neuer-weg weg (first die-kanten))
     (intern (rest die-kanten) start)))
  (intern (nachfolge-kanten (second weg) kanten) (first (reverse weg)))
  )
;Test:
;(zum-start '(99 P H I))
    
;;; ===== Hilfsfunktion expandiere =====
; expandiert die vorhandenen Wege mit den zulaessigen Fortsetzungen
(define
  (expandiere wege)
  (if
    (null? wege) '()
    (append
     (alle-nachfolge-wege (first wege) kanten)
     (expandiere (rest wege)))))
; Test:
;(expandiere '((99 P H I) (96 K H I)))

;;; ===== Suchfunktion Erarbeitung =====
;;; -----------------------------------------
;;; Schritt 1: Kopf
(define
  (suche start anzahl wege bester)
  'fehlt
  )
; Test:
;(suche 'I (anzahl-knoten kanten) '((0 I)) '(+inf.0 I))

;;; Schritt 2: kein Weg mehr vorhanden
;;; besten zurueckgeben (ggf. Misserfolgsfall)
(define
  (suche start anzahl wege bester)
  (cond
    ((null? wege)
     bester)
    (else
     'fehlt)
  ))
; Test:
;(suche 'I (anzahl-knoten kanten) '() '(+inf.0 I))
;(suche 'I (anzahl-knoten kanten) '((0 I)) '(+inf.0 I))

;;; Schritt 3: aktueller Weg enthaelt alle Orte und die
; Nachfolgestaedte enthalten den Startort
; => eine Loesung gefunden -> Vergleich mit bester
(define
  (suche start anzahl wege bester)
  ;(ausgabe bester)
  (cond
    ((null? wege)
     bester)
    ((and
      (= (knoten-im-weg (first wege)) anzahl)
      (member start (nachbar-staedte (second (first wege)) kanten)))
     (suche
      start
      anzahl
      (rest wege)
      (besserer-weg
       bester
       (zum-start (first wege)))))
    (else
     'fehlt)
  ))
; Test:
;(suche 'I (anzahl-knoten kanten) '() '(+inf.0 I))
;(suche 'I (anzahl-knoten kanten) '((0 I)) '(+inf.0 I))
;(suche 'I 3 '((99 P H I)) '(+inf.0 I))

;;; Schritt 4: aktueller Weg enthaelt alle Orte aber die
; Nachfolgestaedte enthalten nicht (!) den Startort
; keine Loesung -> mit naechstem Weg weiter suchen
(define
  (suche start anzahl wege bester)
  ;(ausgabe bester)
  (cond
    ((null? wege)
     bester)
    ((and
      (= (knoten-im-weg (first wege)) anzahl)
      (member start (nachbar-staedte (second (first wege)) kanten)))
     (suche
      start
      anzahl
      (rest wege)
      (besserer-weg
       bester
       (zum-start (first wege)))))
    ((= (knoten-im-weg (first wege)) anzahl)
     (suche
      start
      anzahl
      (rest wege)
      bester))
    (else
     'fehlt)
  ))
; Test:
;(suche 'I (anzahl-knoten kanten) '() '(+inf.0 I))
;(suche 'I (anzahl-knoten kanten) '((0 I)) '(+inf.0 I))
;(suche 'I 3 '((99 P H I)) '(+inf.0 I))
;(suche 'I 3 '((96 K H I)) '(+inf.0 I))

;;; Schritt 5: weiter in der Breite suchen,
; also die vorhandenen Wege expandieren
(define
  (suche start anzahl wege bester)
  ;(ausgabe bester)
  (cond
    ((null? wege)
     bester)
    ((and
      (= (knoten-im-weg (first wege)) anzahl)
      (member start (nachbar-staedte (second (first wege)) kanten)))
     (suche
      start
      anzahl
      (rest wege)
      (besserer-weg
       bester
       (zum-start (first wege)))))
    ((= (knoten-im-weg (first wege)) anzahl)
     (suche
      start
      anzahl
      (rest wege)
      bester))
    (else
     (suche start anzahl (expandiere wege) bester))
  ))
; Test:
(suche 'I (anzahl-knoten kanten) '() '(+inf.0 I))
(suche 'I (anzahl-knoten kanten) '((0 I)) '(+inf.0 I))
(suche 'I 3 '((99 P H I)) '(+inf.0 I))
(suche 'I 3 '((96 K H I)) '(+inf.0 I))

